"""Calculate earth motion around sun"""

# ----------------------------- License information --------------------------

# This file is part of the helisol python package.
# Copyright (C) 2023 Olivier Vincent

# The helisol package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The helisol package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the helisol python package.
# If not, see <https://www.gnu.org/licenses/>


import numpy as np

from .general import CONSTANTS
from .general import Angle, Time, Distance
from .general import AngleFromDegrees, AngleFromSeconds, AngleFromRadians
from .general import sin, cos, tan


perturb_planet_coeffs = CONSTANTS['planet perturbation coefficients']
perturb_planet_longitude = CONSTANTS['planet perturbation longitude']
perturb_planet_radius = CONSTANTS['planet perturbation radius']


class EarthOrbit:
    """Store general characteristics of earth orbit"""

    def __init__(self, utc_time=None):
        """Init earth orbit object from specific date/time.

        Parameters
        ----------
        utc_time: datetime or str (default None, i.e. current time)

        Examples
        --------
        EarthOrbit('9am')
        EarthOrbit(utc_time='June 10 10:08:44')
        """
        self.time = Time(utc_time=utc_time)

    def __repr__(self):
        year = self.time.utc.date().year
        return f'Earth Orbit in {year}'

    @property
    def excentricity(self):
        """Excentricity of earth's orbit"""
        return 0.016709 - 0.000042 * self.time.julian_centuries

    @property
    def axial_tilt(self):
        """Earth's axial tilt (epsilon)"""
        t = self.time.julian_centuries
        eps0 = AngleFromDegrees(23 + 26 / 60 + (21.5 - 46.8 * t) / 3600)
        d_eps = AngleFromSeconds(9.2 * cos(self.nutation))  # tilt nutation
        return eps0 + d_eps

    @property
    def spring_longitude(self):
        """Longitude between perigee and spring"""
        a = CONSTANTS['average motion coefficients']
        t = self.time.julian_centuries
        gamma0 = sum([(a['L'][i] - a['M'][i]) * t**i for i in range(3)])
        return AngleFromDegrees(gamma0)

    @property
    def nutation(self):
        """Ω"""
        t = self.time.julian_centuries
        nut0, nut1 = CONSTANTS['nutation coefficients']
        return AngleFromDegrees(nut0 + nut1 * t)

    @property
    def correc_nutation(self):
        """Δψ"""
        Ω = self.nutation
        return AngleFromSeconds(-17.2 * sin(Ω))

    @property
    def correc_aberration(self):
        """Δaberr"""
        return AngleFromSeconds(-20.5)

    def correc_planets_angle(self, name):
        """name is one of the Meeus coefficient name ('A', 'B', etc.)."""
        t = self.time.julian_centuries
        coeffs = perturb_planet_coeffs[name]
        a = 0
        for i, coeff in enumerate(coeffs):
            a += coeff * t**i
        return AngleFromDegrees(a)

    @property
    def correc_planets(self):
        """Perturbations from planets, moon, etc. on longitude"""
        perturb = AngleFromDegrees()
        for coeff_name, (ampl, func) in perturb_planet_longitude.items():
            ag = self.correc_planets_angle(coeff_name)
            perturb += AngleFromDegrees(ampl * getattr(ag, func)())
        return perturb

    @property
    def correc_planets_distance(self):
        """Perturbations from planets, moon, etc. on earth-sun distance"""
        perturb = 0
        for coeff_name, (ampl, func) in perturb_planet_radius.items():
            ag = self.correc_planets_angle(coeff_name)
            perturb += ampl * getattr(ag, func)()
        return Distance(au=perturb)


class Earth:

    def __init__(self, utc_time=None):
        """Init earth object from specific date/time.

        Parameters
        ----------
        utc_time: datetime or str (default None, i.e. current time)

        Examples
        --------
        Earth('9am')
        Earth(utc_time='June 10 10:08:44')
        """
        self.update(utc_time=utc_time)

    def update(self, utc_time=None):
        self.time = Time(utc_time=utc_time)
        self.orbit = EarthOrbit(utc_time=utc_time)

    def __repr__(self):
        date = self.time.utc.date()
        time = self.time.utc.time().strftime("%H:%M:%S")
        return f'Earth on {date} at {time} (UTC)'

    @property
    def average_motion(self):
        """Average motion with respect to perigee"""
        a = CONSTANTS['average motion coefficients']
        t = self.time.julian_centuries
        m_deg = sum([a['M'][i] * t**i for i in range(3)])
        m = AngleFromDegrees(m_deg)
        m.minus_pi_to_pi()
        return m

    def anomaly(self, iteration=5):
        """Iterative way of calculating excentric anomaly"""
        m0 = self.average_motion
        if iteration == 0:
            return m0
        else:
            u = self.anomaly(iteration=iteration - 1)
            e = self.orbit.excentricity
            m_rad = m0.radians + e * sin(u)
            return AngleFromRadians(m_rad)

    @property
    def true_anomaly(self):
        """True anomaly (nu)"""
        u = self.anomaly(iteration=CONSTANTS['anomaly iterations'])
        e = self.orbit.excentricity
        return 2 * Angle.arctan(np.sqrt((1 + e) / (1 - e)) * tan(u / 2))

    @property
    def longitude(self):
        """Longitude (λ) with respect to spring"""
        λ0 = self.orbit.spring_longitude
        λs = λ0 + self.true_anomaly
        return λs

    @property
    def apparent_longitude(self):
        """Apparent longitude (including correction from nutation, aberration, planets etc.)"""
        λs = self.longitude
        Δψ = self.orbit.correc_nutation
        Δaberr = self.orbit.correc_aberration
        Δplanets = self.orbit.correc_planets
        return λs + Δψ + Δaberr + Δplanets

    @property
    def distance(self):
        e = self.orbit.excentricity
        nu = self.true_anomaly
        R = 1.0000002 * (1 - e**2) / (1 + e * cos(nu))
        ΔR = self.orbit.correc_planets_distance
        return Distance(au=R) + ΔR



"""General functions and tools for the helisol package."""

# ----------------------------- License information --------------------------

# This file is part of the helisol python package.
# Copyright (C) 2023 Olivier Vincent

# The helisol package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The helisol package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the helisol python package.
# If not, see <https://www.gnu.org/licenses/>


# ================================= Imports ==================================


import datetime
import math
from dateutil.parser import parse
from functools import total_ordering
import numpy as np


# ================================ CONSTANTS =================================


CONSTANTS = {}

CONSTANTS['reference time'] = datetime.datetime(2000, 1, 1, 12, 0)

a = {}  # Coefficients for average motion
a['L'] = 280.46646, 36000.76983, 0.0003032   # J. MEEUS SAF
a['M'] = 357.52911, 35999.05029, -0.0001537  # (degrees)

CONSTANTS['average motion coefficients'] = a
CONSTANTS['nutation coefficients'] = (125.04, -1934.136)

CONSTANTS['anomaly iterations'] = 3
CONSTANTS['sunset iterations'] = 2


perturb_planet_coeffs = {'A': (351.52, 22518.443),              # Venus 1
                         'B': (253.14, 45036.886),              # Venus 2
                         'C': (157.23, 32964.467),              # Jupiter
                         'D': (297.85, 445267.112),   # Moon
                         'E': (252.08, 20.190),                   # Long period
                         'H': (353.40, 65928.7155)}              # H

perturb_planet_longitude = {'A': (134e-5, 'cos'),
                            'B': (153e-5, 'cos'),
                            'C': (200e-5, 'cos'),
                            'D': (180e-5, 'sin'),
                            'E': (196e-5, 'sin')}

perturb_planet_radius = {'A': (5.43e-6, 'sin'),
                         'B': (15.75e-6, 'sin'),
                         'C': (16.27e-6, 'sin'),
                         'D': (30.76e-6, 'cos'),
                         'H': (9.27e-6, 'sin')}

CONSTANTS['planet perturbation coefficients'] = perturb_planet_coeffs
CONSTANTS['planet perturbation longitude'] = perturb_planet_longitude
CONSTANTS['planet perturbation radius'] = perturb_planet_radius

astronomical_unit = 149_597_870_700  # in meters


# ============================== MISC. classes ===============================


@total_ordering
class Angle:
    """Store angles and retrieve them in degrees, radians or hours"""
    def __init__(self,
                 degrees=0, minutes=0, seconds=0,
                 hms=None,
                 radians=None):
        """Input angle in degrees (+ minutes/seconds), or in radians, or
        in time-like format (h, m, s).

        Notes:
        - By default, value is considered to be decimal degrees, e.g. Angle(10.7)
        - degrees, minutes, seconds can be floats and are added
          (but it's preferrable to use ints when using  deg, min, sec instead
           of decimal degrees)
        - same for h, m, s
        - ValueError raised if input has mixed units between °, hms, rad
        """
        # ATTENTION the nature of the tests below can significantly reduce
        # speed of some operations, e.g. calculating sunsets, sunrises etc.
        test_input_deg = any([bool(x) for x in (degrees, minutes, seconds)])
        test_input_hms = (hms is not None)
        test_input_rad = (radians is not None)
        if sum([test_input_deg, test_input_hms, test_input_rad]) > 1:
            raise ValueError('Cannot specify angle in more than one set of units')

        if test_input_rad:
            self._degrees = radians * 180 / np.pi
        elif test_input_hms:
            h, m, s = hms
            self._degrees = 15 * (h + m / 60 + s / 3600)
        else:
            self._degrees = degrees + minutes / 60 + seconds / 3600

    def __repr__(self):
        sign = '-' if self.sign < 0 else ''
        degrees = abs(int(self.degrees))
        minutes = abs(self.minutes)
        seconds = abs(self.seconds)
        h, m, s = [abs(x) for x in self.hms]
        a = "helisol.Angle"
        b = f"""\n{sign}{degrees}°{minutes}'{seconds:.2f}" """
        c = f"[{sign}{h}h{m}m{s:.2f}s]\n"
        d = f'{self.degrees} [°]\n'
        e = f'{self.radians} [rad]'
        return a + b + c + d + e

    @property
    def sign(self):
        """1 if angle is positive or zero, -1 if negative"""
        if self.degrees == 0:
            return 1
        else:
            return (int(self.degrees > 0)) - (int(self.degrees < 0))

    def __eq__(self, other):
        return (self.degrees == other.degrees)

    def __lt__(self, other):
        return (self.degrees < other.degrees)

    def __add__(self, other):
        return Angle(degrees=self.degrees + other.degrees)

    def __sub__(self, other):
        return Angle(degrees=self.degrees - other.degrees)

    def __neg__(self):
        return Angle(degrees=-self.degrees)

    def __mul__(self, other):
        return Angle(degrees=self.degrees * other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return Angle(degrees=self.degrees / other)

    @property
    def radians(self):
        return self._degrees * np.pi / 180  # much faster than np.radians() for single values

    @property
    def degrees(self):
        return self._degrees

    @property
    def decimal_minutes(self):
        return 60 * (self.degrees % self.sign)

    @property
    def minutes(self):
        return int(self.decimal_minutes)

    @property
    def seconds(self):
        """Only for info, not settable"""
        return 60 * (self.decimal_minutes % self.sign)

    @property
    def hms(self):
        """Tuple hours (int), minutes (int), seconds (float)"""
        h_decimal = self.degrees / 15
        h = int(h_decimal)
        m_decimal = 60 * (h_decimal % self.sign)
        m = int(m_decimal)
        s = 60 * (m_decimal % self.sign)
        return (h, m, s)

    def sin(self):
        return math.sin(self.radians)

    def cos(self):
        return math.cos(self.radians)

    def tan(self):
        return math.tan(self.radians)

    def cotan(self):
        return 1 / math.tan(self.radians)

    def minus_pi_to_pi(self):
        """Angle modulo 2*pi, between -pi and pi"""
        self._degrees = math.atan2(self.sin(), self.cos()) * 180 / np.pi

    @classmethod
    def arcsin(cls, value):
        return Angle(radians=math.asin(value))

    @classmethod
    def arccos(cls, value):
        return Angle(radians=math.acos(value))

    @classmethod
    def arctan(cls, value):
        return Angle(radians=math.atan(value))

    @classmethod
    def arctan2(cls, x1, x2):
        return Angle(radians=math.atan2(x1, x2))


@total_ordering
class AngleArray:
    """Store arrays of angles and retrieve them in degrees or radians."""
    def __init__(self, degrees=None, minutes=None, seconds=None, radians=None):
        """Input angle in degrees (+ minutes/seconds), or in radians.

        Notes:
        - By default, value is considered to be decimal degrees
        - degrees, minutes, seconds can be floats and are added
          (but it's preferrable to use ints when using  deg, min, sec instead
           of decimal degrees)
        - ValueError raised if input has mixed units between ° and rad
        """
        test_input_deg = any([x is not None for x in (degrees, minutes, seconds)])
        test_input_rad = (radians is not None)
        if sum([test_input_deg, test_input_rad]) > 1:
            raise ValueError('Cannot specify angle in more than one set of units')

        if radians is not None:
            self._degrees = np.degrees(radians)
            return

        val1 = np.array(degrees) if degrees is not None else 0
        val2 = np.array(minutes) / 60 if minutes is not None else 0
        val3 = np.array(seconds) / 3600 if seconds is not None else 0
        self._degrees = val1 + val2 + val3

    def __repr__(self):
        a = "helisol.AngleArray"
        b = f'{self.degrees} [°]\n'
        c = f'{self.radians} [rad]'
        return a + b + c

    def __eq__(self, other):
        return (self.degrees == other.degrees)

    def __lt__(self, other):
        return (self.degrees < other.degrees)

    def __add__(self, other):
        return AngleArray(degrees=self.degrees + other.degrees)

    def __sub__(self, other):
        return AngleArray(degrees=self.degrees - other.degrees)

    def __neg__(self):
        return AngleArray(degrees=-self.degrees)

    def __mul__(self, other):
        return AngleArray(degrees=self.degrees * other)

    def __rmul__(self, other):
        return AngleArray(degrees=other * self.degrees)

    def __truediv__(self, other):
        return AngleArray(degrees=self.degrees / other)

    @property
    def radians(self):
        return np.radians(self._degrees)

    @property
    def degrees(self):
        return self._degrees

    def sin(self):
        return np.sin(self.radians)

    def cos(self):
        return np.cos(self.radians)

    def tan(self):
        return np.tan(self.radians)

    def cotan(self):
        return 1 / np.tan(self.radians)

    def minus_pi_to_pi(self):
        """Array modulo 2*pi, between -pi and pi"""
        self._degrees = np.degrees(np.arctan2(self.sin(), self.cos()))

    @classmethod
    def arcsin(cls, value):
        return AngleArray(radians=np.arcsin(value))

    @classmethod
    def arccos(cls, value):
        return AngleArray(radians=np.arccos(value))

    @classmethod
    def arctan(cls, value):
        return AngleArray(radians=np.arctan(value))

    @classmethod
    def arctan2(cls, x1, x2):
        return AngleArray(radians=np.arctan2(x1, x2))


class AngleFromDegrees(Angle):
    """Shortcut for faster execution of Angle when known to instantiate from degrees"""

    def __init__(self, value=0):
        self._degrees = value


class AngleFromMinutes(Angle):
    """Shortcut for faster execution of Angle when known to instantiate from arcminutes"""

    def __init__(self, value=0):
        self._degrees = value / 60


class AngleFromSeconds(Angle):
    """Shortcut for faster execution of Angle when known to instantiate from arcseconds"""

    def __init__(self, value=0):
        self._degrees = value / 3600


class AngleFromRadians(Angle):
    """Shortcut for faster execution of Angle when known to instantiate from radians"""

    def __init__(self, value=0):
        self._degrees = value * 180 / np.pi


# ===================== Convenience functions on angles ======================


def sin(angle):
    return angle.sin()


def cos(angle):
    return angle.cos()


def tan(angle):
    return angle.tan()


def cotan(angle):
    return angle.cotan()


# =========================== Distance management ============================


@total_ordering
class Distance:

    def __init__(self, m=0, km=None, au=None):
        """Init distance with meters, kilometers or astronomical units.

        Data can be single values or arrays.
        """
        try:
            test_input_m = bool(abs(m) > 0)
        except ValueError:  # if array is passed, consider it as an input
            test_input_m = True
        test_input_km = (km is not None)
        test_input_au = (au is not None)
        if sum([test_input_m, test_input_km, test_input_au]) > 1:
            raise ValueError('Cannot specify angle in more than one set of units')

        if test_input_km:
            self.km = km
        elif test_input_au:
            self.au = au
        else:
            self.m = m

    def __repr__(self):
        return f'helisol.Distance\n{self.m}[m]\n{self.km}[km]\n{self.au}[A.U.]'

    def __eq__(self, other):
        return (self.m == other.m)

    def __lt__(self, other):
        return (self.m < other.m)

    def __add__(self, other):
        return Distance(m=self.m + other.m)

    def __sub__(self, other):
        return Distance(m=self.m - other.m)

    def __neg__(self):
        return Distance(m=-self.m)

    def __mul__(self, other):
        return Distance(m=self.m * other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return Distance(m=self.m / other)

    @property
    def m(self):
        return self._m

    @m.setter
    def m(self, value):
        self._m = value
        self._km = value / 1000
        self._au = value / astronomical_unit

    @property
    def km(self):
        return self._km

    @km.setter
    def km(self, value):
        self._km = value
        self._m = value * 1000
        self._au = self._m / astronomical_unit

    @property
    def au(self):
        return self._au

    @au.setter
    def au(self, value):
        self._au = value
        self._m = value * astronomical_unit
        self._km = self._m / 1000


# =========================== Date/Time management ===========================


class Time:
    """Store time in various useful formats"""

    def __init__(self, utc_time=None, fraction_of_day=None):
        """Init time object.

        Parameters
        ----------
        - utc_time: datetime or str (default None, i.e. current time)

        - fraction_of_day: if specified, overrides time from the given fraction
                           of day (0.5 for Noon) [keeps input date]
        """
        self.utc = self._parse_time(utc_time)
        self._update()

        if fraction_of_day is not None:
            self.fraction_of_day = fraction_of_day

    def __repr__(self):
        return str(self.utc)

    def _update(self):
        """Recalculate quantities if UTC time has changed."""
        self.elapsed = self.utc - CONSTANTS['reference time']
        self.days = self.elapsed.total_seconds() / (24 * 3600)
        self.julian_years = self.days / 365.25
        self.julian_centuries = self.julian_years / 100

    def _parse_time(self, utc_time=None):
        """Return a datetime object from user input"""
        if utc_time is None:
            return datetime.datetime.utcnow()
        else:
            return parse(str(utc_time), yearfirst=True)  # str is in case a datetime object is passed

    @property
    def fraction_of_day(self):
        """return 0 for midnight, 0.5 for noon"""
        time = self.utc.time()
        return (time.hour * 3600 + time.minute * 60 + time.second) / (24 * 3600)

    @fraction_of_day.setter
    def fraction_of_day(self, value):
        """Inverse function for fraction_of_day().

        Kind of using hack from
        https://stackoverflow.com/questions/656297/python-time-timedelta-equivalent
        """
        Δt = datetime.timedelta(days=1) * value
        date = self.utc.date()
        midnight = datetime.time()
        self.utc = datetime.datetime.combine(date, midnight) + Δt
        self._update()

    def rounded_to(self, unit='second'):
        """Return rounded version of self.

        Parameters
        ----------
        - unit (str): 'second' or 'minute'
        """
        if unit == 'second':
            old_time = self.utc
            new_time = old_time.replace(microsecond=0)
            if old_time.microsecond >= 5e5:
                new_time += datetime.timedelta(seconds=1)

        elif unit == 'minute':
            old_time = self.rounded_to('second').utc
            new_time = old_time.replace(second=0)
            if old_time.second >= 30:
                new_time += datetime.timedelta(minutes=1)

        return Time(utc_time=new_time)


# ================================ Refraction ================================


def refraction_saemundsson_math(true_height):
    """Refraction angle from true height (Saemundsson 1986).

    Uses math.tan (faster for single values)

    Parameters
    ----------
    true_height: angle in degrees

    Output
    ------
    Refraction angle in arc-minutes
    """
    y = true_height + (10.3 / (true_height + 5.11))
    return 1.02 / math.tan(y * np.pi / 180)


def refraction_saemundsson_np(true_height):
    """Refraction angle from true height (Saemundsson 1986)

    Uses np.tan (for arrays)

    Parameters
    ----------
    true_height: angle in degrees

    Output
    ------
    Refraction angle in arc-minutes
    """
    y = true_height + (10.3 / (true_height + 5.11))  # in degrees
    return 1.02 / np.tan(y * np.pi / 180)


def refraction_bennett_np(apparent_height=None):
    """Refraction angle from apparent height (Bennett 1982)

    Parameters
    ----------
    apparent_height: angle in degrees

    Output
    ------
    Refraction angle in arc-minutes
    """
    y0 = apparent_height + 7.31 / (apparent_height + 4.4)  # in degrees
    return 1 / np.tan(y0 * np.pi / 180)


def refraction(true_height=None, apparent_height=None):
    """Refraction angle, either from true height or apparent height.

    Parameters
    ----------
    - true_height: helisol.Angle
    - apparent_height: helisol.Angle

    If both true_height and apparent_height are provided, apparent_height
    is ignored.

    True height formula is from Saemundsson 1986
    Apparent height formula is from Bennett 1982

    Normally valid for 1010 bars of pressure and 10°C of temperature.

    Examples
    --------
    refraction(true_height=Angle(0))
    refraction(apparent_height=Angle(0))
    refraction(Angle(23))                          # true height by default
    refraction(true_height=Angle(degrees=23))
    refraction(apparent_height=Angle(minutes=66))
    """
    if true_height is not None:
        return AngleFromMinutes(refraction_saemundsson_np(true_height.degrees))
    else:
        return AngleFromMinutes(refraction_bennett_np(apparent_height.degrees))



"""Calculate sun position"""

# ----------------------------- License information --------------------------

# This file is part of the helisol python package.
# Copyright (C) 2023 Olivier Vincent

# The helisol package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The helisol package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the helisol python package.
# If not, see <https://www.gnu.org/licenses/>


from copy import copy
import numpy as np

from .general import CONSTANTS, Angle, Time
from .general import AngleFromDegrees
from .general import refraction, refraction_saemundsson_math, sin, cos, tan
from .earth import Earth
from .locations import Location


class Sun:
    """Sun description irrespective of location on earth"""

    def __init__(self, utc_time=None):
        """Init sun object at specific date/time.

        Parameters
        ----------
        - utc_time: datetime or str (default None, i.e. current time)

        Examples
        --------
        Sun('9am')
        Sun(utc_time='June 10 10:08:44')
        """
        self.update(utc_time=utc_time)

    def __repr__(self):
        date = self.time.utc.date()
        time = self.time.utc.time().strftime("%H:%M:%S")
        a = f'Sun on {date} at {time} (UTC)\n'

        ẟ = self.declination.degrees
        alpha = self.right_ascension.degrees
        eqt = self.equation_of_time.degrees
        b = f'Declination [{ẟ:.1f}°], Right Asc. [{alpha:.1f}°], EQT [{eqt:.1f}°]'

        return a + b

    def update(self, utc_time=None):
        """Update at current time (default) or given UTC time."""
        self.time = Time(utc_time=utc_time)
        self.earth = Earth(utc_time=utc_time)

    @property
    def angular_diameter(self):
        d = 1_392_684
        l = self.earth.distance.km
        return Angle.arctan(d / l)

    @property
    def right_ascension(self):
        ε = self.earth.orbit.axial_tilt
        λapp = self.earth.apparent_longitude
        return Angle.arctan2(cos(ε) * sin(λapp), cos(λapp))

    @property
    def declination(self):
        ε = self.earth.orbit.axial_tilt
        λapp = self.earth.apparent_longitude
        return Angle.arcsin(sin(λapp) * sin(ε))

    @property
    def equation_of_time(self):
        alpha = self.right_ascension
        gamma0 = self.earth.orbit.spring_longitude
        m = self.earth.average_motion
        Δaberr = self.earth.orbit.correc_aberration
        Δψ = self.earth.orbit.correc_nutation
        ε = self.earth.orbit.axial_tilt

        eqt = (alpha - gamma0 - m - Δaberr - Δψ * cos(ε))
        eqt.minus_pi_to_pi()

        return eqt


class SunObservation:
    """Observation of sun at specific location on earth."""

    def __init__(self, location=None, utc_time=None):
        """Init sun observation object from specific location and date/time.

        Parameters
        ----------
        - location: Location name from JSON database,
                    [or] custom Location object
                    [or] iterable (latitude, longitude) in degrees
                    if None (default), use default location

        - utc_time: datetime or str (default None, i.e. current time)

        Examples
        --------
        loc = (42.4, -76.5)
        Sun(loc, '9am')
        Sun(location=loc, utc_time='June 10 10:08:44')
        """
        self.location = Location.parse(location)
        self.latitude, self.longitude = [AngleFromDegrees(x) for x in self.location.coords]
        self.update(utc_time=utc_time)

    def update(self, utc_time=None):
        """Update at current time (default) or given UTC time."""
        self.time = Time(utc_time=utc_time)
        self.earth = Earth(utc_time=utc_time)
        self.sun = Sun(utc_time=utc_time)

    def __repr__(self):

        lat = self.latitude.degrees
        long = self.longitude.degrees
        date = self.time.utc.date()
        time = self.time.utc.time().strftime("%H:%M:%S")
        a = f'Sun observation from ({lat}°, {long}°) on {date} at {time} (UTC)'

        height = self.height.degrees
        app_h = self.apparent_height.degrees
        azimuth = self.azimuth.degrees
        b = f'\nAzimuth [{azimuth:.1f}°], Height [{height:.1f}°], App. Height [{app_h:.1f}°]'

        moments = self.sunrise.utc, self.noon.utc, self.sunset.utc
        sunrise, noon, sunset = [m.strftime("%H:%M:%S") for m in moments]
        c = f'\nSunrise {sunrise} Noon {noon} Sunset {sunset}'

        return a + b + c

    @property
    def hourly_angle(self):
        x = self.time.fraction_of_day
        a = AngleFromDegrees(360 * (x - 0.5))
        return a - self.sun.equation_of_time + self.longitude

    @property
    def height(self):
        """Height of sun above horizon."""
        ẟ = self.sun.declination
        lat = self.latitude
        H = self.hourly_angle
        return Angle.arcsin(sin(ẟ) * sin(lat) + cos(ẟ) * cos(lat) * cos(H))

    @property
    def apparent_height(self):
        """Apparent height including refraction"""
        h = self.height
        return h + refraction(true_height=h)

    @property
    def azimuth(self):
        """Azimuth of sun with respect to South"""
        ẟ = self.sun.declination
        H = self.hourly_angle
        lat = self.latitude
        a = cos(ẟ) * sin(H)
        b = cos(ẟ) * cos(H) * sin(lat) - sin(ẟ) * cos(lat)
        return Angle.arctan2(a, b)

    # --------------------- Sunrise, solar noon, sunset ----------------------

    def _noon(self):
        """Raw solar noon, calculated for current time."""
        eqt = self.sun.equation_of_time
        long = self.longitude
        noon_frac = (eqt.radians - long.radians) / (2 * np.pi) + 0.5
        return Time(utc_time=self.time.utc.date(), fraction_of_day=noon_frac)

    def _sunrise(self):
        """Sunrise"""
        ẟ = self.sun.declination
        lat = self.latitude
        noon_frac = self._noon().fraction_of_day
        rise_frac = noon_frac - np.arccos(-tan(ẟ) * tan(lat)) / (2 * np.pi)
        return Time(utc_time=self.time.utc.date(), fraction_of_day=rise_frac)

    def _sunset(self):
        """Sunset"""
        set_frac = 2 * self._noon().fraction_of_day - self._sunrise().fraction_of_day
        return Time(utc_time=self.time.utc.date(), fraction_of_day=set_frac)

    def _converge_to_event(self, event='noon', iteration=CONSTANTS['sunset iterations']):
        """Iterative way of conver ging sun towards sunrise, noon, or sunset"""
        raw_event_name = '_' + event
        if iteration == 0:
            return self
        else:
            old_obs = self._converge_to_event(event, iteration=iteration - 1)
            event_time = getattr(old_obs, raw_event_name)()
            return SunObservation(old_obs.location, utc_time=event_time)

    @property
    def noon(self):
        return self._converge_to_event(event='noon')._noon()

    @property
    def sunrise(self):
        return self._converge_to_event(event='sunrise')._sunrise()

    @property
    def sunset(self):
        return self._converge_to_event(event='sunset')._sunset()

    def _actual_event(self,
                      event='sunrise',
                      refract=True,
                      point='center',
                      obstacle=None,
                      precision=0.01,
                      print_details=False):
        """Find actual moment of sunset or sunrise (with refraction / obstacles).
        (iteratively)

        Parameters
        ----------
        - event: 'sunrise' or 'sunset'
        - refract: (default: True): take into account refraction or not
        - point: consider 'center', 'top', or 'bottom' of the sun
        - obstacle: angular height of obstacle masking the sun (Angle object)
                    or function of the azimuth (°) returning an angle (°).
                    By default, no obstacle (obstacle angle of 0°).
        - precision: which (angular) tolerance to consider matching heights
                     NOTE: cannot be lower than 0.0002° (or 0.7 arcseconds or
                     0.05 time-seconds) and even sometimes 0.002° (for some
                     reason I do not completely understand yet) because one
                     runs into precision limitations for floats in Python.
                     Default: 0.01°.
        - print_details: print info on the iteration / convergence process
        """
        consider_obstacle = bool(obstacle)
        if consider_obstacle:
            try:
                obstacle(0)
            except TypeError:
                obstacle_height = obstacle      # obstacle is an angle
                obstacle_is_func = False
            else:
                obstacle_height = obstacle      # obstacle is func(azimuth)
                obstacle_is_func = True

        coeffs = {'sunrise': -1, 'sunset': 1}
        point_coeff = {'center': 0, 'bottom': -1, 'top': 1}
        c = coeffs[event]
        p = point_coeff[point]

        # Initial guess: ~7 minutes before sunrise (or after sunset) at horizon
        f0 = getattr(self, event).fraction_of_day + c * 5e-3
        obs_search = copy(self)

        def match_heights(f):
            """Function that returns 0 sun point matches given height"""
            time = Time(obs_search.time, fraction_of_day=f)
            obs_search.update(utc_time=time)

            h = obs_search.height.degrees
            az = obs_search.azimuth.degrees

            if point in ('top', 'bottom'):
                diam = obs_search.sun.angular_diameter.degrees
                h += p * diam / 2

            if refract:
                h += refraction_saemundsson_math(h) / 60  # minutes to degrees

            if consider_obstacle:
                if obstacle_is_func:
                    h -= obstacle_height(az)
                else:
                    h -= obstacle_height

            return h

        def mvtime(f0, step=1e-4, max_it=1e4):
            """Iterative search."""
            i = 0
            f = f0
            m0 = match_heights(f0)
            direction = np.sign(m0) * c

            stop = False
            found = False
            max_iterations = False

            while not (found or stop or max_iterations):
                m = match_heights(f)
                if abs(m) < precision:
                    found = True
                elif np.sign(m) == - np.sign(m0):
                    stop = True
                elif i > max_it:
                    max_iterations = True
                else:
                    i += 1
                    f += direction * step

            return {'value': f,
                    'found': found,
                    'stopped': stop,
                    'iterations': i,
                    'max iterations': max_iterations,
                    'residual': m}

        def manage_result(results):
            results['iterations'] = total_iterations
            if print_details:
                print(results)
            f = results['value']
            return Time(utc_time=self.time, fraction_of_day=f)

        f = f0
        total_iterations = 0

        for step in (1e-2, 1e-3, 1e-4, 1e-5, 1e-6):
            results = mvtime(f, step=step)
            f = results['value']
            total_iterations += results['iterations']
            if results['found']:
                return manage_result(results)
        else:
            if results['found']:
                return manage_result(results)
            else:
                msg = f'Impossible to converge {event} search within tolerance. '
                msg += f'This might be because required precision of {precision}° is too high '
                msg += 'or because the input obstacle height is larger than the maximum height of the sun. '
                msg += 'Please try first lower precision and/or lower obstacle height. '
                results['iterations'] = total_iterations
                msg += str(results)
                raise RuntimeError(msg)

    def actual_sunrise(self, *args, point='top', **kwargs):
        """Find actual moment of sunrise (with refraction / obstacles), iteratively.

        Parameters
        ----------
        - refract: (default: True): take into account refraction or not
        - point: consider 'center', 'top', or 'bottom' of the sun
        - obstacle: angular height of obstacle masking the sun (Angle object)
                    or function of the azimuth (°) returning an angle (°).
                    By default, no obstacle (obstacle angle of 0°).
        - precision: which (angular) tolerance to consider matching heights
                     NOTE: cannot be lower than 0.0002° (or 0.7 arcseconds or
                     0.05 time-seconds) and even sometimes 0.002° (for some
                     reason I do not completely understand yet) because one
                     runs into precision limitations for floats in Python.
                     Default: 0.01°.
        - print_details: print info on the iteration / convergence process
        """
        return self._actual_event(event='sunrise', *args, point=point, **kwargs)

    def actual_sunset(self, *args, point='top', **kwargs):
        """Find actual moment of sunrise (with refraction / obstacles), iteratively.

        Parameters
        ----------
        - refract: (default: True): take into account refraction or not
        - point: consider 'center', 'top', or 'bottom' of the sun
        - obstacle: angular height of obstacle masking the sun (Angle object)
                    or function of the azimuth (°) returning an angle (°).
                    By default, no obstacle (obstacle angle of 0°).
        - precision: which (angular) tolerance to consider matching heights
                     NOTE: cannot be lower than 0.0002° (or 0.7 arcseconds or
                     0.05 time-seconds) and even sometimes 0.002° (for some
                     reason I do not completely understand yet) because one
                     runs into precision limitations for floats in Python.
                     Default: 0.01°.
        - print_details: print info on the iteration / convergence process
        """
        return self._actual_event(event='sunset', *args, point=point, **kwargs)



"""Management of locations"""

# ----------------------------- License information --------------------------

# This file is part of the helisol python package.
# Copyright (C) 2023 Olivier Vincent

# The helisol package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The helisol package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the helisol python package.
# If not, see <https://www.gnu.org/licenses/>

import helisol
from copy import copy
import json
from json import JSONDecodeError
from dataclasses import dataclass
from pathlib import Path
from typing import Tuple

from .config import CONFIG

LOCATIONS_FOLDER = Path(helisol.__file__).parent.parent / 'locations'
LOCATIONS_FILES = 'perso.json', 'global.json'

@dataclass
class Location:
    name: str
    coords: Tuple
    elevation: int = None

    @classmethod
    def parse(cls, user_input):
        """Generate Location object from user input."""
        try:
            user_input.coords  # Location object or equivalent
        except AttributeError:
            pass
        else:
            return user_input

        if user_input is None:
            return Location.load(CONFIG['default location'])

        elif type(user_input) is str:
            return Location.load(user_input)

        try:
            lat, long = [float(x) for x in user_input]
        except (TypeError, ValueError):
            raise ValueError(f'Unknown location: {user_input}')
        else:
            return Location(name='[Unspecified]', coords=(lat, long))

    @classmethod
    def load(cls, name):
        """Create location object from name.

        Parameter
        ---------
        name (str): name of location as stored in the JSON database

        Example
        -------
        location = Location.load('Home')
        """
        for filename in LOCATIONS_FILES:
            file = LOCATIONS_FOLDER / filename
            all_data = cls._from_json(file)
            try:
                data = all_data[name]
                return cls(name=name, **data)
            except KeyError:
                pass
        else:
            raise ValueError(f'Location name "{name}" not found in JSON database')

    def save(self, kind='perso'):
        """Save location data into JSON database.

        Parameter
        ---------
        kind (str): 'perso' (default: non-shared locations), or 'global' (shared)

        Example
        -------
        location.save('global')
        """
        filename = kind + '.json'
        file = LOCATIONS_FOLDER / filename

        data = self._from_json(file)  # load existing data to add to it

        props = copy(vars(self))
        name = props.pop('name')
        data[name] = props

        self._to_json(file, data)

    def remove(cls, name, kind='perso'):
        """Remove location entry from JSON database"""
        filename = kind + '.json'
        file = LOCATIONS_FOLDER / filename
        data = cls._from_json(file)
        try:
            data.pop(name)
        except KeyError:
            raise ValueError(f'No entry named {name} in {filename} database')
        else:
            cls._to_json(file, data)

    @staticmethod
    def _from_json(file):
        """Load python data (dict or list) from json file"""
        try:
            with open(file, 'r', encoding='utf8') as f:
                data = json.load(f)
        except (FileNotFoundError, JSONDecodeError):  # no existing file or data
            data = {}
        return data

    @staticmethod
    def _to_json(file, data):
        """Load python data (dict or list) from json file"""
        with open(file, 'w', encoding='utf8') as f:
            json.dump(data, f, indent=4, ensure_ascii=False)



"""Manage command line parsing for the helisol module."""

# ----------------------------- License information --------------------------

# This file is part of the helisol python package.
# Copyright (C) 2023 Olivier Vincent

# The helisol package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The helisol package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the helisol python package.
# If not, see <https://www.gnu.org/licenses/>


import argparse


descr = "Calculate sun position, sunset, sunrise, and solar noon."

parser = argparse.ArgumentParser(description=descr,
                                 formatter_class=argparse.RawTextHelpFormatter)

msg = "Input location"

# The nargs='?' is to have a positional argument with a default value
parser.add_argument('location', type=str, nargs='?', help=msg)
args = parser.parse_args()



"""Helisol package init."""

# ----------------------------- License information --------------------------

# This file is part of the helisol python package.
# Copyright (C) 2023 Olivier Vincent

# The helisol package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The helisol package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the helisol python package.
# If not, see <https://www.gnu.org/licenses/>


from .general import Time, Distance, Angle, AngleArray
from .general import AngleFromRadians, AngleFromDegrees, AngleFromMinutes, AngleFromSeconds
from .general import sin, cos, tan, cotan, refraction
from .general import astronomical_unit
from .earth import Earth
from .sun import Sun, SunObservation
from .tables import generate_table, sunset_table, extend_table
from .locations import Location

# from importlib.metadata import version (only for python 3.8+)
from importlib_metadata import version


__version__ = version('helisol')
__author__ = 'Olivier Vincent'
__license__ = 'GNU GPLv3'



"""Generate data tables."""

# ----------------------------- License information --------------------------

# This file is part of the helisol python package.
# Copyright (C) 2023 Olivier Vincent

# The helisol package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The helisol package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the helisol python package.
# If not, see <https://www.gnu.org/licenses/>


import datetime
import pandas as pd
from oclock import parse_time

from .general import Time, AngleFromDegrees
from .sun import SunObservation
from .locations import Location


column_names = {
    'azimuth': 'Azimuth (°)',
    'height': 'Height (°)',
    'apparent_height': 'Apparent Height (°)',
    'sun.declination': 'Declination (°)',
    'sun.right_ascension': 'Right Ascension (°)',
    'sun.equation_of_time': 'Equation of Time (°)',
}


def _generate_times(start, end, interval):
    """Generate list of helisol.Time objects between specified times."""
    t1 = Time(start)
    t2 = Time(end)
    dt = parse_time(interval)

    times = []
    t = t1.utc
    while t <= t2.utc:
        times.append(Time(t))
        t += dt

    return times


def _get_value(obj, name):
    """Get value of object property from its name, including subnames.

    (e.g. _get_value(observation, sun.declination))
    """
    for attrib in name.split('.'):
        obj = getattr(obj, attrib)
    return obj


def generate_table(location, start, end, interval, columns=column_names):
    """Generate table with solar data at regular intervals between two dates.

    Parameters
    ----------
     - location: Location name from JSON database,
                    [or] custom Location object
                    [or] iterable (latitude, longitude) in degrees
                    if None (default), use default location

    - start: datetime, helisol.Time object, str or other datetime information (UTC)

    - end: datetime, helisol.Time object, str or other datetime information (UTC)

    - interval: str in the format 'hh:mm:ss', including '::10' for 10 secs

    - columns (optional): specify which properties and columns will be included
                          (dict with sun attribute name as key and name of
                           corresponding column as value)

    Output
    ------
    Pandas DataFrame

    Examples
    --------
    generate_table(location=(47, 2),
                   start='March 1 12:00',
                   end='March 31 12:00',
                   interval='24::',
                   columns={'azimuth': 'Azimuth (°)', 'height': 'Height (°)'})
    """
    location = Location.parse(location)
    times = _generate_times(start=start, end=end, interval=interval)

    data = {}
    data['Date'] = [time.utc.date() for time in times]
    data['Time (UTC)'] = [time.utc.time() for time in times]

    observations = [SunObservation(location=location, utc_time=time)
                    for time in times]

    for ppty, name in columns.items():
        data[name] = [_get_value(obs, ppty).degrees
                      for obs in observations]

    return pd.DataFrame(data)


def sunset_table(location, start, end, refract=True, point='top', rounding='second'):
    """Create able with sunrise, noon and sunsets between specified dates.

    NOTE: it can take a little while for large datasets
          (~ 4-5 seconds for a month of data).

    Parameters
    ----------
     - location: Location name from JSON database,
                    [or] custom Location object
                    [or] iterable (latitude, longitude) in degrees
                    if None (default), use default location

    - start: datetime, helisol.Time object, str or other datetime information (UTC)

    - end: datetime, helisol.Time object, str or other datetime information (UTC)

    - refract: Take refraction or not into account.

    - point: which point of sun to consider ('top', 'bottom', 'center')

    - rounding: 'second' or 'minute'

    Output
    ------
    Pandas DataFrame
    """
    times = _generate_times(start=start, end=end, interval='24::')

    data = {}
    data['Date'] = [time.utc.date() for time in times]

    observations = [SunObservation(location=location, utc_time=time)
                    for time in times]

    for ppty in 'sunrise', 'noon', 'sunset':
        name = ppty.capitalize()
        ll = []
        for obs in observations:
            if ppty == 'noon' or (not refract and point == 'center'):
                time = _get_value(obs, ppty)
                ftime = time
            else:
                ppty_func = 'actual_' + ppty
                time_func = _get_value(obs, ppty_func)
                ftime = time_func(refract=refract, point=point, precision=0.002)
            ll.append(ftime.rounded_to(rounding).utc.time())
        data[name] = ll

    return pd.DataFrame(data)


def extend_table(data,
                 location,
                 date_column='Date',
                 time_column='Time (UTC)',
                 columns=column_names):
    """Takes an existing table with date / time columns and adds columns with
    the corresponding solar data.

    Parameters
    ----------
    - data: pandas DataFrame

     - location: Location name from JSON database,
                    [or] custom Location object
                    [or] iterable (latitude, longitude) in degrees
                    if None (default), use default location

    - date_column: name of the column containing dates in the pandas dataframe

    - time_column: name of the column containing time in the pandas dataframe

    - columns (optional): specify which properties and columns will be included
                          (dict with sun attribute name as key and name of
                           corresponding column as value)

    Example
    -------
    data = pd.read_excel('Data.xls')
    extend_table(data, location=(47, 2))
    """
    def _calculate_angle(ppty, row):
        utc_time = datetime.datetime.combine(row[date_column], row[time_column])
        obs = SunObservation(location=location, utc_time=utc_time)
        return _get_value(obs, ppty).degrees

    for ppty, name in columns.items():
        calc = lambda row: _calculate_angle(ppty, row)
        data[name] = data.apply(calc, axis=1)



"""Test helisol module with pytest"""

from pathlib import Path
import pandas as pd
import numpy as np

import helisol
from helisol import Sun, Time, SunObservation, Distance, astronomical_unit
from helisol import Angle, AngleArray, AngleFromDegrees, AngleFromRadians
from helisol import AngleFromMinutes, AngleFromSeconds
from helisol import refraction

DATA_FOLDER = Path(helisol.__file__).parent.parent / 'data'


# ================================ Misc tools ================================


def predict_event(row, event='sunrise'):
    obs = SunObservation((47, 2), utc_time=row['Date'])
    if event in ('sunrise', 'sunset'):
        ppty = 'actual_' + event
        event_time = getattr(obs, ppty)(point='center', precision=0.002)
    else:
        event_time = getattr(obs, event)
    return event_time.utc.time()


def event_diff(row, event='sunrise'):
    column_eph = event.capitalize()
    column_th = column_eph + ' (predicted)'
    event_eph = Time(row[column_eph]).utc
    event_th = Time(row[column_th]).utc
    return (event_th - event_eph).total_seconds()


# ================================== Tests ===================================


def test_angle():
    """Test basic angle operation"""
    a = Angle(30)
    b = Angle(radians=(np.pi * 60 / 180))
    c = Angle(hms=(2, 0, 0))
    assert round((a + b - c).cos(), 3) == 0.5


def test_specialized_angle():
    """Test faster Angle implementations"""
    a = AngleFromDegrees(30)
    b = AngleFromRadians(np.pi / 6)
    c = AngleFromMinutes(1.5)
    d = 90 * AngleFromSeconds(1)
    assert round((a - b + c - d).seconds, 2) == 0


def test_angle_array():
    """Test AngleArray class."""
    a = AngleArray(degrees=[44.5, 2.5, 3], minutes=[30, 30, 0])
    assert round(a.tan()[0], 6) == 1


def test_distance():
    """Test Distance class"""
    d = Distance(au=1)
    assert round(d.km * 1000 / astronomical_unit, 6) == 1


def test_sun():
    """Test sun object"""
    sun = Sun('Aug 10, 2023, 12:00')
    assert round(sun.declination.degrees, 1) == 15.6
    assert round(sun.equation_of_time.degrees, 1) == 1.4
    assert round(sun.right_ascension.degrees, 1) == 140.1
    assert round(sun.angular_diameter.degrees, 2) == 0.53


def test_sun_observation():
    """Test calculation of sun height and azimuth"""
    obs = SunObservation(location=(42.4, -76.5))
    obs.update(utc_time='Jan 6, 2023, 4:25:03pm')
    assert round(obs.azimuth.degrees, 2) == -11.84
    assert round(obs.height.degrees, 2) == 24.24


def test_sunset():
    """Test calculation of sunset time"""
    obs = SunObservation(location=(42.4, -76.5), utc_time='2023-1-6, 16:25:03')
    assert obs.sunset.utc.hour == 21
    assert obs.sunset.utc.minute == 43


def test_actual_sunset():
    obs = SunObservation(location=(47, 2), utc_time='2023-03-15')
    sunset = obs.actual_sunset(point='center', precision=0.002).rounded_to('minute')
    assert sunset.utc.hour == 17
    assert sunset.utc.minute == 56


def test_refraction_true_height():
    """Test calculation of refraction from true height"""
    r1 = refraction(true_height=Angle(degrees=0))
    r2 = refraction(Angle(minutes=-34))
    assert r1.minutes == 28
    assert r2.minutes == 34


def test_refraction_apparent_height():
    """Test calculation of refraction from true height"""
    r1 = refraction(apparent_height=Angle(degrees=0))
    r2 = refraction(apparent_height=Angle(minutes=34))
    assert r1.minutes == 34
    assert r2.minutes == 28


def test_ephemerides_march2023():
    """Check that predicted values coincide with Ephemerides"""
    file = DATA_FOLDER / 'Ephemerides_March2023_47N_2E.tsv'
    df = pd.read_csv(file, sep='\t')
    for event in ('sunrise', 'noon', 'sunset'):
        pred = lambda x: predict_event(x, event)
        diff = lambda x: event_diff(x, event)
        column_pred = event.capitalize() + ' (predicted)'
        column_diff = event.capitalize() + ' (diff)'
        df[column_pred] = df.apply(pred, axis=1)
        df[column_diff] = df.apply(diff, axis=1)
    assert (df['Noon (diff)'].abs() < 1).all()
    assert (df['Sunset (diff)'].abs() < 60).all()
    assert (df['Sunrise (diff)'].abs() < 60).all()


def test_ephemerides_aug2023():
    file = DATA_FOLDER / 'Ephemerides_Aug2023.tsv'
    df = pd.read_csv(file, sep='\t')

    # Right_ascension --------------------------------------------------------

    def analyze_asc(row):
        utc_time = row['Date'] + ' ' + row['Time']
        sun = Sun(utc_time=utc_time)
        h, m, s = [int(x) for x in row['Right Ascension'].split(':')]
        asc_eph = Angle(hms=(h, m, s))
        asc_th = sun.right_ascension
        return (asc_eph.degrees - asc_th.degrees) * 240  # in time-seconds

    df['Right Ascension (diff)'] = df.apply(analyze_asc, axis=1)
    max_dev_asc_time_s = df['Right Ascension (diff)'].abs().describe()['max']

    # Declination ------------------------------------------------------------

    def analyze_decl(row):
        utc_time = row['Date'] + ' ' + row['Time']
        sun = Sun(utc_time=utc_time)
        deg, mm, ss = [int(x) for x in row['Declination'].split(':')]
        decl_eph = Angle(degrees=deg, minutes=mm, seconds=ss)
        decl_th = sun.declination
        return (decl_eph.degrees - decl_th.degrees) * 3600  # in arc-seconds

    df['Declination (diff)'] = df.apply(analyze_decl, axis=1)
    max_dev_decl_arc_s = df['Declination (diff)'].abs().describe()['max']

    # Equation of time -------------------------------------------------------

    def analyze_eqt(row):
        utc_time = row['Date'] + ' ' + row['Time']
        sun = Sun(utc_time=utc_time)
        m, s = [int(x) for x in row['Equation of Time'].split(':')]
        eqt_eph = Angle(hms=(0, m, s))
        eqt_th = sun.equation_of_time
        return (eqt_eph.degrees - eqt_th.degrees) * 240

    df['Equation of Time (diff)'] = df.apply(analyze_eqt, axis=1)
    max_dev_eqt_time_s = df['Equation of Time (diff)'].abs().describe()['max']

    # Earth-sun distance -----------------------------------------------------

    def analyze_dist(row):
        utc_time = row['Date'] + ' ' + row['Time']
        sun = Sun(utc_time=utc_time)
        return sun.earth.distance.au

    df['Distance to Earth (predicted)'] = df.apply(analyze_dist, axis=1)
    df['Distance to Earth (diff, km)'] = (df['Distance to Earth'] - df['Distance to Earth (predicted)']) * astronomical_unit / 1000
    df.plot(y=['Distance to Earth', 'Distance to Earth (predicted)'])
    max_dev_dist_km = df['Distance to Earth (diff, km)'].describe()['max']

    # Final tests ------------------------------------------------------------

    assert max_dev_asc_time_s < 1     # Right asc. error < 1 seconds (in time)
    assert max_dev_decl_arc_s < 2.5   # Declination error < 2.5 arc-seconds
    assert max_dev_eqt_time_s < 1     # EQT error < 1 seconds (in time)
    assert max_dev_dist_km < 1600    # Distance earth-sun, error < 10000 km
